#!/bin/sh -e

# dash on ubuntu <= 12.04 does not support [^...]
# mozart does not support [[:space:]]
# artpec-4 does not support grep -m1

parse_dst()
{
  dst=${1#<*>}
  dst=${dst#*[0-9]}
  dst=${dst#[0-9]}
  dst=${dst#:[0-9][0-9]}
}

parse_posix()
{
  local dst p1 p2
  parse_dst "$1"
  p1=${1%$dst}
  p2=${dst%%,*}

  start_end=${1#$p1}
  start_end=${start_end#$p2}

  off1=${p1#<*>}
  off1=${off1##*[^0-9:-]}
  off2=${p2#<*>}
  off2=${off2##*[^0-9:-]}
}

format_name()
{
  local name="$1"
  local posix="$2"
  local off1 off2 start_end
  local sign='-'
  local oh
  local om=0
  local dst

  parse_posix "$posix"
  oh=${off1%:*}
  if [ "$oh" != "$off1" ]; then
    om=${off1#*:}
  fi
  if [ "$oh" != "${oh#-}" ]; then
    sign='+'
    oh=${oh#-}
  fi

  name=${name#(*) }
  name=${name% (+DST)}
  [ -n "$start_end" ] && dst=' (+DST)' || dst=''
  printf "(UTC%c%02d:%02d) %s%s\n" "$sign" "$oh" "$om" "$name" "$dst"
}

build_i610()
{
  local name iana posix line dst
  local off1 off2 start_end
  local any="\([A-Z]\+\|<[^>]*>\)"
  
  while IFS='|' read name iana posix; do
    if ! line=$(grep -m1 "|$posix$" /etc/sysconfig/tzlist.conf); then
      parse_posix "$posix"
      dst="$off2$start_end"
      if [ -n "$dst" ]; then
        dst="$any$dst"
      fi
      line=$(grep -m1 "|$any$off1$dst$" /etc/sysconfig/tzlist.conf) || true
    fi
    if [ -n "$line" ]; then
      posix=${line##*|}
      iana=${line#*|}
      iana=${iana#*|}
      iana=${iana%%|*}
      if [ "$iana" != "${iana#Etc/}" ]; then
        posix='<'${iana#Etc/}'>'${posix#GMT}
      fi
      echo "$name|$iana|$posix"
    fi
  done < "$zonesfile.orig" > "$zonesfile"
}

build_i825()
{
  local name iana posix line dst
  local off1 off2 start_end
  local any="\([A-Z]\+\|<[^>]*>\)"

  while IFS='|' read name iana posix; do
    if ! line=$(grep -m1 " $posix$" /opt/ad/etc/conf/tz-list-uniq.conf); then
      parse_posix "$posix"
      dst="$off2$start_end"
      if [ -n "$dst" ]; then
        dst="$any$dst"
      fi
      line=$(grep -m1 " $any$off1$dst$" /opt/ad/etc/conf/tz-list-uniq.conf) || true
    fi
    if [ -n "$line" ]; then
      posix=${line##* }
      iana=${line%% *}
      iana=${iana#/usr/share/zoneinfo/}
      echo "$name|$iana|$posix"
    fi
  done < "$zonesfile.orig" > "$zonesfile"
}

list_timezones()
{
  local name iana posix dst
  while IFS='|' read name iana posix; do
    format_name "$name" "$posix"
  done < "$zonesfile"
}

write_tzfile()
{
  local posix=$1
  local z3='\0\0\0'
  local z7=$z3$z3'\0'
  local z18=$z7$z7$z3'\0'
  local z25=$z18$z7
  local z29=$z25$z3'\0'
  printf "TZif2$z18\2$z3\2$z7\1$z3\2$z25"
  printf "TZif2$z18\2$z3\2$z7\1$z3\2$z29"
  printf "\n$posix\n"
}

find_timezone()
{
  local line

  name="$1"
  name=${name#(*) }
  if line=$(grep -e "^$name|" -e "|$name|" "$zonesfile" | head -n1); then
    name=${line%%|*}
    iana=${line#*|}
    iana=${iana%|*}
    posix=${line##*|}
  elif [ -f "/usr/share/zoneinfo/$name" ]; then
    iana=$name
    parse_tzfile "/usr/share/zoneinfo/$iana"
    line=$(grep "|$posix$" "$zonesfile" || true)
    name=${line%%|*}
  else
    name=''
    iana=''
    posix=''
  fi
}

parse_tzfile()
{
  local tzfile="$1"
  local has_newline=$(( $(tail -c1 "$tzfile" 2> /dev/null | wc -l) ))
  if [ "$has_newline" = "1" ]; then
    posix=$(tail -n1 "$tzfile")
  else
    local abbrev=$(TZ="$tzfile" date +'%Z')
    local offset=$(TZ="$tzfile" date +'%z')
    local sign=${offset%[0-9]*}
    [ "$sign" = "-" ] && sign='' || sign='-'
    local hour=${offset#[+-]}
    hour=${hour%[0-9][0-9]}
    hour=${hour#0}
    local min=${offset#[+-]}
    min=${min#[0-9][0-9]}
    offset="$sign$hour"
    if [ "$min" != '00' ]; then
      offset="$offset:$min"
    elif [ "$hour" = '0' -a "$sign" = '-' ]; then
      offset='0'
    fi
    posix="$abbrev$offset"
  fi
}

get_timezone()
{
  local name iana posix expected

  if [ -e "$namefile" ]; then
    name=$(cat "$namefile")
    find_timezone "$name"
  fi

  if [ "$os" != 'vivotek' ]; then
    expected=$posix
    parse_tzfile /etc/localtime
    if [ "$posix" != "$expected" ]; then
      name=$(grep "|$posix$" "$zonesfile" || true)
      name=${name%%|*}
    fi
  fi

  if [ -z "$posix" ]; then
    name='UTC'
    posix='UTC0'
  fi
  if [ -z "$name" ]; then
    name='Custom'
  fi
  format_name "$name" "$posix"
  echo "$posix"
}

set_timezone()
{
  local name iana posix desired_posix dst

  find_timezone "$1"
  if [ -z "$name" ]; then
    echo 'timezone not found.'
    exit 1
  fi

  local desired_posix="$posix"

  case "$os" in
  axis)
    parse_dst "$posix"
    [ -n "$dst" ] && dst='yes' || dst='no'
    parhandclient set time.posixtimezone "$posix"
    parhandclient set time.dst.enabled "$dst"
    # axis will accept any posix string
    # wait for new setting
    parhandclient get time.posixtimezone
    ;;
  i610)
    local prev=$(paramclient -get -raw system.datetime.tz.name)
    if [ "$posix" != "${posix#<}" ]; then
      posix=GMT${posix#<*>}
    fi
    paramclient -set "system.datetime.tz.name=$posix"
    if /etc/init.d/timezone start; then
      paramclient -set "system.datetime.tz.posixrule=$posix"
    else
      paramclient -set "system.datetime.tz.name=$prev"
      exit 1
    fi
    ;;
  i825)
    if [ "$posix" != "${posix#<}" ]; then
      posix=${posix#<}
      posix=${posix%%>*}
    fi
    /opt/ad/bin/tzset "$posix"
    ;;
  vivotek)
    # cannot write to /etc, so core always sets the TZ env variable
    ;;
  *)
    local tzfile="/usr/share/zoneinfo/$iana"
    if [ -f "$tzfile" ]; then
      parse_tzfile "$tzfile"
      if [ "$posix" != "$desired_posix" ]; then
        tzfile=''
      fi
    else
      tzfile=''
    fi
    if [ -z "$tzfile" ]; then
      tzfile="/usr/share/zoneinfo/Custom"
      write_tzfile "$desired_posix" > "$tzfile"
    fi
    ln -fs "$tzfile" /etc/localtime
    ;;
  esac

  if [ -f "/usr/share/zoneinfo/$iana" ]; then
    echo "$iana" > "$namefile.exacq"
  else
    echo "$name" > "$namefile.exacq"
  fi
  mv "$namefile.exacq" "$namefile"

  echo "$desired_posix"
}

fix_i610()
{
  # i610 may forget the timezone after reboot
  local posix
  posix=$(paramclient -get -raw system.datetime.tz.name)
  write_tzfile "$posix" > /var/config/localtime.exacq
  ln -fs /var/config/localtime.exacq /var/config/localtime
}

start_ntp()
{
  if ! pidof ntpd > /dev/null && [ -x /etc/init.d/ntp ]; then
    for server in $1; do
      ntpdate -b "$server"
      break
    done
    /etc/init.d/ntp start
  fi
}

get_ntp()
{
  local enabled disabled name value ignore name2 server
  local start="$1"

  while read name value ignore; do
    name2=${name#[#]}
    if [ "$name2" = 'server' -o "$name2" = 'pool' ]; then
      if [ "$value" != 'localhost' -a "$value" = "${value#127.}" ]; then
        if [ "$name" = "$name2" ]; then
          enabled="$enabled $value"
        else
          disabled="$disabled $value"
        fi
      fi
    fi
  done <<EOF
$(sed 's/^#[[:space:]]\+/#/' /etc/ntp.conf)
EOF

  if [ -n "$enabled" ]; then
    echo "enabled"
    for server in $enabled; do
      echo $server
    done
  else
    echo "disabled"
    for server in $disabled; do
      echo $server
    done
  fi

  if [ $start -eq 1 ]; then
    start_ntp "$enabled"
  fi
}

set_ntp()
{
  local enable="$1"
  local servers="$2"
  local driftfile='/var/lib/ntp/ntp.drift'
  local server
  local start="$1"

  /etc/init.d/ntp stop

  if [ -f '/etc/ntp.drift' ]; then
    driftfile='/etc/ntp.drift'
  fi

  {
    for server in $servers; do
      if [ "$enable" = '1' ]; then
        echo "server $server"
      else
        echo "#server $server"
      fi
    done
    cat <<EOF
server 127.127.1.0
fudge 127.127.1.0 stratum 13
driftfile $driftfile
restrict -4 default kod notrap nomodify nopeer noquery
restrict -6 default kod notrap nomodify nopeer noquery
restrict 127.0.0.1
EOF
  } > /etc/ntp.conf.exacq

  mv /etc/ntp.conf.exacq /etc/ntp.conf
  if [ $start -eq 1 ]; then
    if [ "$enable" = '1' ]; then
      start_ntp "$servers"
    else
      start_ntp
    fi
  fi
}

usage()
{
  echo "Usage:"
  echo "$0 [list-timezones|get-timezone]"
  echo "$0 get-ntp [-start-ntp]"
  echo "$0 set-time <timestamp>"
  echo "$0 set-timezone <name|iana>"
  echo "$0 set-ntp [-enable|-disable] [-server <server>]... [-start-ntp]"
  exit 1
}

os='linux'
namefile='/etc/timezone'
zonesfile="${0%/*}/../zones.txt"

if type parhandclient > /dev/null 2> /dev/null; then
  os='axis'
elif type paramclient > /dev/null 2> /dev/null; then
  os='i610'
  namefile='timezone'
  if [ ! -e "$zonesfile" ]; then
    build_i610
  fi
  if [ ! -e /var/config/localtime -a "$UID" -eq 0 ]; then
    fix_i610
  fi
elif [ -x /opt/ad/bin/tzset ]; then
  os='i825'
  if [ ! -e "$zonesfile" ]; then
    build_i825
  fi
elif [ -e /mnt/auto/CF/vadp ]; then
  os='vivotek'
  namefile='timezone'
fi

case "$1" in
  init)
    ;;
  set-time)
    time=${2%%[^0-9]*}
    if [ -z "$time" ]; then
      usage
    fi
    if [ "$os" = 'linux' ]; then
      # must open /dev/null or date will fail
      date -s "@$time" > /dev/null
      # #15676 - if systemd is installed, it disables hwclock
      # if ntp is also installed, systemd-timesyncd disables itself,
      # and the hardware clock does not get updated if the ntp client
      # does not synchronize to an external server
      hwclock --systohc
    else
      date -s "$time" -D '%s'
    fi
    ;;
  list-timezones)
    list_timezones
    ;;
  get-timezone)
    get_timezone
    ;;
  set-timezone)
    if [ -z "$2" ]; then
      usage
    fi
    set_timezone "$2"
    ;;
  get-ntp)    
    shift 1
    start_ntp=0
    if [ -n "$1" ]; then
        if [ "$1" = "-start-ntp" ]; then
            start_ntp=1
        else
            usage
        fi
    fi    
    get_ntp "$start_ntp"
    ;;
  set-ntp)
    shift 1
    enable=1
    servers=''
    start_ntp=0
    while [ -n "$1" ]; do
      case "$1" in
        -enable)
          enable=1
          ;;
        -disable)
          enable=0
          ;;
        -server)
          if [ -$z "$2" ]; then
            usage
          fi
          servers="$servers $2"
          shift 1
          ;;
        -start-ntp)
          start_ntp=1
          ;;
        *)
          usage
          ;;
      esac
      shift 1
    done
    if [ "$enable" = '1' -a -z "$servers" ]; then
      usage
    fi
    set_ntp "$enable" "$servers" "$start_ntp"
    ;;
  *)
    usage
    ;;
esac
